/* DebugEventsPlugin
 *
 * Simple NPAPI plugin for testing the browser/plugin interface, especially
 * to see what kinds of events the browser passes to the plugin.
 *
 * The original template was the NetscapeCocoaPlugin sample that comes with
 * the WebKit source distro (in WebKitExamplePlugins).  QuickDraw support
 * was scavanged from various places in the Mozilla tree -- particularly the
 * Null/Default plugin.
 */

#import "npapi.h"
#import "npfunctions.h"
#import "npruntime.h"
#import "npupp.h"

#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>

#define EXPORT __attribute__((visibility("default")))

#define SUPPORT_CORE_GRAHPICS "supportCoreGraphics"
#define SUPPORT_COCOA "supportCocoa"

#ifdef __ppc__

// glue for mapping outgoing Macho function pointers to TVectors
struct TFPtoTVGlue{
    void* glue[2];
};

struct pluginFuncsGlueTable {
    TFPtoTVGlue     newp;
    TFPtoTVGlue     destroy;
    TFPtoTVGlue     setwindow;
    TFPtoTVGlue     newstream;
    TFPtoTVGlue     destroystream;
    TFPtoTVGlue     asfile;
    TFPtoTVGlue     writeready;
    TFPtoTVGlue     write;
    TFPtoTVGlue     print;
    TFPtoTVGlue     event;
    TFPtoTVGlue     urlnotify;
    TFPtoTVGlue     getvalue;
    TFPtoTVGlue     setvalue;

    TFPtoTVGlue     shutdown;
} gPluginFuncsGlueTable;

static inline void* SetupFPtoTVGlue(TFPtoTVGlue* functionGlue, void* fp)
{
    functionGlue->glue[0] = fp;
    functionGlue->glue[1] = 0;
    return functionGlue;
}

#define PLUGIN_TO_HOST_GLUE(name, fp) (SetupFPtoTVGlue(&gPluginFuncsGlueTable.name, (void*)fp))

// glue for mapping netscape TVectors to Macho function pointers
struct TTVtoFPGlue {
    uint32 glue[6];
};

struct netscapeFuncsGlueTable {
    TTVtoFPGlue             geturl;
    TTVtoFPGlue             posturl;
    TTVtoFPGlue             requestread;
    TTVtoFPGlue             newstream;
    TTVtoFPGlue             write;
    TTVtoFPGlue             destroystream;
    TTVtoFPGlue             status;
    TTVtoFPGlue             uagent;
    TTVtoFPGlue             memalloc;
    TTVtoFPGlue             memfree;
    TTVtoFPGlue             memflush;
    TTVtoFPGlue             reloadplugins;
    TTVtoFPGlue             getJavaEnv;
    TTVtoFPGlue             getJavaPeer;
    TTVtoFPGlue             geturlnotify;
    TTVtoFPGlue             posturlnotify;
    TTVtoFPGlue             getvalue;
    TTVtoFPGlue             setvalue;
    TTVtoFPGlue             invalidaterect;
    TTVtoFPGlue             invalidateregion;
    TTVtoFPGlue             forceredraw;
    TTVtoFPGlue             pushpopupsenabledstate;
    TTVtoFPGlue             poppopupsenabledstate;
} gNetscapeFuncsGlueTable;

static void* SetupTVtoFPGlue(TTVtoFPGlue* functionGlue, void* tvp)
{
    static const TTVtoFPGlue glueTemplate = { 0x3D800000, 0x618C0000, 0x800C0000, 0x804C0004, 0x7C0903A6, 0x4E800420 };

    memcpy(functionGlue, &glueTemplate, sizeof(TTVtoFPGlue));
    functionGlue->glue[0] |= ((UInt32)tvp >> 16);
    functionGlue->glue[1] |= ((UInt32)tvp & 0xFFFF);
    ::MakeDataExecutable(functionGlue, sizeof(TTVtoFPGlue));
    return functionGlue;
}

#define HOST_TO_PLUGIN_GLUE(name, fp) (SetupTVtoFPGlue(&gNetscapeFuncsGlueTable.name, (void*)fp))

#else

#define PLUGIN_TO_HOST_GLUE(name, fp) (fp)
#define HOST_TO_PLUGIN_GLUE(name, fp) (fp)

#endif /* __ppc__ */

// Browser function table
static NPNetscapeFuncs browser = {0};

// Structure for per-instance storage
typedef struct PluginObject
{
    NPP npp;

    NPWindow window;
    uint16 mode;

    NSString *debugInfo;

    NPBool supportsCoreGraphics;
    NPBool supportsCocoa;

    NPBool hasFocus;

    GrafPtr savedPortQD;
    RgnHandle savedClipQD;
    short savedPortTopQD;
    short savedPortLeftQD;
} PluginObject;

NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16 mode, int16 argc, char* argn[], char* argv[], NPSavedData* saved);
NPError NPP_Destroy(NPP instance, NPSavedData** save);
NPError NPP_SetWindow(NPP instance, NPWindow* window);
NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16* stype);
NPError NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason);
int32 NPP_WriteReady(NPP instance, NPStream* stream);
int32 NPP_Write(NPP instance, NPStream* stream, int32 offset, int32 len, void* buffer);
void NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname);
void NPP_Print(NPP instance, NPPrint* platformPrint);
int16 NPP_HandleEvent(NPP instance, void* event);
void NPP_URLNotify(NPP instance, const char* URL, NPReason reason, void* notifyData);
NPError NPP_GetValue(NPP instance, NPPVariable variable, void *value);
NPError NPP_SetValue(NPP instance, NPNVariable variable, void *value);

// Mach-o entry points
extern "C" {
EXPORT int main(NPNetscapeFuncs* browserFuncs, NPPluginFuncs* pluginFuncs, NPP_ShutdownUPP* unloadUpp);
EXPORT NPError NP_Initialize(NPNetscapeFuncs *browserFuncs);
EXPORT NPError NP_GetEntryPoints(NPPluginFuncs *pluginFuncs);
EXPORT void NP_Shutdown(void);
}

// Needed for old fashioned browsers ... like FF2 and Opera.
int main(NPNetscapeFuncs* browserFuncs, NPPluginFuncs* pluginFuncs, NPP_ShutdownUPP* unloadUpp)
{
    NSLog(@"DebugEventsPlugin: main(): browserFuncs 0x%.8X, pluginFuncs 0x%.8X, unloadUpp 0x%.8X",
          (unsigned) browserFuncs, (unsigned) pluginFuncs, (unsigned) unloadUpp);
    if (!browserFuncs || !pluginFuncs || !unloadUpp) {
        NSLog(@"DebugEventsPlugin: main() error return");
        return NPERR_INVALID_FUNCTABLE_ERROR;
    }

    browser.version            = browserFuncs->version;
    browser.size               = browserFuncs->size;
    browser.getvalue           = (NPN_GetValueUPP)HOST_TO_PLUGIN_GLUE(getvalue, browserFuncs->getvalue);
    browser.setvalue           = (NPN_SetValueUPP)HOST_TO_PLUGIN_GLUE(setvalue, browserFuncs->setvalue);
    browser.invalidaterect     = (NPN_InvalidateRectUPP)HOST_TO_PLUGIN_GLUE(invalidaterect, browserFuncs->invalidaterect);

    pluginFuncs->version       = 11;
    pluginFuncs->size          = sizeof(pluginFuncs);
    pluginFuncs->newp          = NewNPP_NewProc(PLUGIN_TO_HOST_GLUE(newp, NPP_New));
    pluginFuncs->destroy       = NewNPP_DestroyProc(PLUGIN_TO_HOST_GLUE(destroy, NPP_Destroy));
    pluginFuncs->setwindow     = NewNPP_SetWindowProc(PLUGIN_TO_HOST_GLUE(setwindow, NPP_SetWindow));
    pluginFuncs->newstream     = NewNPP_NewStreamProc(PLUGIN_TO_HOST_GLUE(newstream, NPP_NewStream));
    pluginFuncs->destroystream = NewNPP_DestroyStreamProc(PLUGIN_TO_HOST_GLUE(destroystream, NPP_DestroyStream));
    pluginFuncs->asfile        = NewNPP_StreamAsFileProc(PLUGIN_TO_HOST_GLUE(asfile, NPP_StreamAsFile));
    pluginFuncs->writeready    = NewNPP_WriteReadyProc(PLUGIN_TO_HOST_GLUE(writeready, NPP_WriteReady));
    pluginFuncs->write         = (NPP_WriteProcPtr)NewNPP_WriteProc(PLUGIN_TO_HOST_GLUE(write, NPP_Write));
    pluginFuncs->print         = NewNPP_PrintProc(PLUGIN_TO_HOST_GLUE(print, NPP_Print));
    pluginFuncs->event         = NewNPP_HandleEventProc(PLUGIN_TO_HOST_GLUE(event, NPP_HandleEvent));
    pluginFuncs->urlnotify     = NewNPP_URLNotifyProc(PLUGIN_TO_HOST_GLUE(urlnotify, NPP_URLNotify));
    pluginFuncs->getvalue      = NewNPP_GetValueProc(PLUGIN_TO_HOST_GLUE(getvalue, NPP_GetValue));
    pluginFuncs->setvalue      = NewNPP_SetValueProc(PLUGIN_TO_HOST_GLUE(setvalue, NPP_SetValue));

    *unloadUpp                 = NewNPP_ShutdownProc(PLUGIN_TO_HOST_GLUE(shutdown, NP_Shutdown));

    return NPERR_NO_ERROR;
}

NPError NP_Initialize(NPNetscapeFuncs* browserFuncs)
{
    NSLog(@"DebugEventsPlugin: NP_Initialize(): browserFuncs 0x%.8X", (unsigned) browserFuncs);
    if (!browserFuncs) {
        NSLog(@"DebugEventsPlugin: NP_Initialize() error return");
        return NPERR_INVALID_FUNCTABLE_ERROR;
    }

    browser.version = browserFuncs->version;
    browser.size = browserFuncs->size;
    browser.getvalue = browserFuncs->getvalue;
    browser.setvalue = browserFuncs->setvalue;
    browser.invalidaterect = browserFuncs->invalidaterect;

    return NPERR_NO_ERROR;
}

NPError NP_GetEntryPoints(NPPluginFuncs* pluginFuncs)
{
    NSLog(@"DebugEventsPlugin: NP_GetEntryPoints(): pluginFuncs 0x%.8X", (unsigned) pluginFuncs);
    if (!pluginFuncs) {
        NSLog(@"DebugEventsPlugin: NP_GetEntryPoints() error return");
        return NPERR_INVALID_FUNCTABLE_ERROR;
    }

    pluginFuncs->version = 11;
    pluginFuncs->size = sizeof(pluginFuncs);
    pluginFuncs->newp = NPP_New;
    pluginFuncs->destroy = NPP_Destroy;
    pluginFuncs->setwindow = NPP_SetWindow;
    pluginFuncs->newstream = NPP_NewStream;
    pluginFuncs->destroystream = NPP_DestroyStream;
    pluginFuncs->asfile = NPP_StreamAsFile;
    pluginFuncs->writeready = NPP_WriteReady;
    pluginFuncs->write = (NPP_WriteProcPtr)NPP_Write;
    pluginFuncs->print = NPP_Print;
    pluginFuncs->event = NPP_HandleEvent;
    pluginFuncs->urlnotify = NPP_URLNotify;
    pluginFuncs->getvalue = NPP_GetValue;
    pluginFuncs->setvalue = NPP_SetValue;

    return NPERR_NO_ERROR;
}

void NP_Shutdown(void)
{
    NSLog(@"DebugEventsPlugin: NP_Shutdown()");
}

static void checkForOverride(char *setting, NPBool *value, NPBool *valueFound,
                             int16 argc, char* argn[], char* argv[])
{
    if (!valueFound)
        return;
    *valueFound = FALSE;
    if (!setting || !value || !argc || !argn || !argv)
        return;

    for (int16 i = 0; i < argc; ++i) {
        if (strcasecmp(setting, argn[i]))
            continue;
        if (!strcasecmp(argv[i], "0") || !strcasecmp(argv[i], "false") || !strcasecmp(argv[i], "no")) {
            *valueFound = TRUE;
            *value = FALSE;
            return;
        }
        if (!strcasecmp(argv[i], "1") || !strcasecmp(argv[i], "true") || !strcasecmp(argv[i], "yes")) {
            *valueFound = TRUE;
            *value = TRUE;
            return;
        }
    }
}

static char *modeString(uint16 mode)
{
    switch (mode) {
        case NP_EMBED: return "NP_EMBED";
        case NP_FULL: return "NP_FULL";
        default: return "Unknown";
    }
}

NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16 mode, int16 argc, char* argn[], char* argv[], NPSavedData* saved)
{
    NSLog(@"DebugEventsPlugin: NPP_New() start: pluginType %s, instance 0x%.8X, mode %s",
           pluginType, (unsigned) instance, modeString(mode));
    if (!instance) {
        NSLog(@"***NULL instance!***");
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    // Create per-instance storage
    PluginObject *obj = (PluginObject *)malloc(sizeof(PluginObject));
    bzero(obj, sizeof(PluginObject));

    obj->npp = instance;
    instance->pdata = obj;

    // Ask the browser if it supports the Cocoa event model
    NPBool supportsCocoa;
    if (browser.getvalue(instance, NPNVsupportsCocoaBool, &supportsCocoa) != NPERR_NO_ERROR)
        supportsCocoa = FALSE;
    // Needed for Opera
    if ((supportsCocoa != TRUE) && (supportsCocoa != FALSE))
        supportsCocoa = FALSE;

    // If the browser does support the Cocoa event model, check if that
    // setting has been overridden.
    if (supportsCocoa) {
        NPBool hasOverride = FALSE;
        NPBool override = TRUE;
        checkForOverride(SUPPORT_COCOA, &override, &hasOverride, argc, argn, argv);
        if (hasOverride)
            supportsCocoa = override;
    }

    obj->supportsCocoa = supportsCocoa;

    // Enable the Cocoa event model if the browser supports it and
    // this hasn't been overridden.
    if (supportsCocoa)
        browser.setvalue(instance, NPPVpluginEventModel, (void *)NPEventModelCocoa);

    // Ask the browser if it supports the CoreGraphics drawing model
    NPBool supportsCoreGraphics;
    if (browser.getvalue(instance, NPNVsupportsCoreGraphicsBool, &supportsCoreGraphics) != NPERR_NO_ERROR)
        supportsCoreGraphics = FALSE;
    // Needed for Opera
    if ((supportsCoreGraphics != TRUE) && (supportsCoreGraphics != FALSE))
        supportsCoreGraphics = FALSE;

    // If the browser does support the CoreGraphics drawing model, check if
    // that setting has been overridden.  But don't do this if we've enabled
    // the Cocoa event model -- Safari (the only browser that currently
    // supports the Cocoa event model) currently crashes if we turn on the
    // Cocoa event model but not the CoreGraphics drawing model.
    if (supportsCoreGraphics && !supportsCocoa) {
        NPBool hasOverride = FALSE;
        NPBool override = TRUE;
        checkForOverride(SUPPORT_CORE_GRAHPICS, &override, &hasOverride, argc, argn, argv);
        if (hasOverride)
            supportsCoreGraphics = override;
    }

    obj->supportsCoreGraphics = supportsCoreGraphics;

    // Enable the CoreGraphics drawing model if the browser supports it and
    // this hasn't been overridden.
    if (supportsCoreGraphics)
        browser.setvalue(instance, NPPVpluginDrawingModel, (void *)NPDrawingModelCoreGraphics);


    obj->mode = mode;

    NSLog(@"DebugEventsPlugin: NPP_New() finish: supportsCoreGraphics %i, supportsCocoa %i",
           supportsCoreGraphics, supportsCocoa);

    return NPERR_NO_ERROR;
}

static NPBool checkInstance(NPP instance)
{
    if (!instance) {
        NSLog(@"***NULL instance!***");
        return FALSE;
    } else if (!instance->pdata) {
        NSLog(@"***Instance not initialized!***");
        return FALSE;
    } else if (instance->pdata == (void *) -1L) {
        NSLog(@"***Instance destroyed!***");
        return FALSE;
    }
    return true;
}

NPError NPP_Destroy(NPP instance, NPSavedData** save)
{
    NSLog(@"DebugEventsPlugin: NPP_Destroy(): instance 0x%.8X", (unsigned) instance);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    // Free per-instance storage
    PluginObject *obj = (PluginObject *) instance->pdata;
    instance->pdata = (void *) -1L; // Mark destroyed

    [obj->debugInfo release];

    free(obj);

    return NPERR_NO_ERROR;
}

NPError NPP_SetWindow(NPP instance, NPWindow* window)
{
    NSLog(@"DebugEventsPlugin: NPP_SetWindow(): instance 0x%.8X, window 0x%.8X, "
           "x %i y %i width %u height %u, clipRect (top %u left %u bottom %u right %u)",
           (unsigned) instance, window ? (unsigned) window->window : 0,
           window ? window->x : 0, window ? window->y : 0,
           window ? window->width : 0, window ? window->height : 0,
           window ? window->clipRect.top : 0, window ? window->clipRect.left : 0,
           window ? window->clipRect.bottom : 0, window ? window->clipRect.right : 0);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    PluginObject *obj = (PluginObject *) instance->pdata;
    obj->window = *window;

    return NPERR_NO_ERROR;
}
 

NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16* stype)
{
    NSLog(@"DebugEventsPlugin: NPP_NewStream(): instance 0x%.8X, stream 0x%.8X, type %s",
           (unsigned) instance, (unsigned) stream, type);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    *stype = NP_ASFILEONLY;
    return NPERR_NO_ERROR;
}

NPError NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason)
{
    NSLog(@"DebugEventsPlugin: NPP_DestroyStream(): instance 0x%.8X, stream 0x%.8X",
           (unsigned) instance, (unsigned) stream);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    return NPERR_NO_ERROR;
}

int32 NPP_WriteReady(NPP instance, NPStream* stream)
{
    NSLog(@"DebugEventsPlugin: NPP_WriteReady(): instance 0x%.8X, stream 0x%.8X",
           (unsigned) instance, (unsigned) stream);
    checkInstance(instance);

    return 0;
}

int32 NPP_Write(NPP instance, NPStream* stream, int32 offset, int32 len, void* buffer)
{
    NSLog(@"DebugEventsPlugin: NPP_Write(): instance 0x%.8X, stream 0x%.8X, offset %i, len %i",
           (unsigned) instance, (unsigned) stream, offset, len);
    checkInstance(instance);

    return 0;
}

void NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname)
{
    NSLog(@"DebugEventsPlugin: NPP_StreamAsFile(): instance 0x%.8X, stream 0x%.8X, fname %s",
           (unsigned) instance, (unsigned) stream, fname);
    checkInstance(instance);
}

void NPP_Print(NPP instance, NPPrint* platformPrint)
{
    NSLog(@"DebugEventsPlugin: NPP_Print(): instance 0x%.8X, mode %s",
           (unsigned) instance, platformPrint ? modeString(platformPrint->mode) : "NULL");
    checkInstance(instance);
}

static NPBool startDrawQD(PluginObject *obj)
{
    if (!obj || obj->supportsCoreGraphics)
        return FALSE;

    NP_Port *npport = (NP_Port *) obj->window.window;
    CGrafPtr ourPort = npport->port;

    if (obj->window.clipRect.left < obj->window.clipRect.right) {
        ::GetPort(&obj->savedPortQD);
        ::SetPort((GrafPtr) ourPort);
        Rect portRect;
        ::GetPortBounds(ourPort, &portRect);
        obj->savedPortTopQD = portRect.top;
        obj->savedPortLeftQD = portRect.left;
        ::GetClip(obj->savedClipQD);

        Rect revealedRectQD;
        revealedRectQD.top = obj->window.clipRect.top + npport->porty;
        revealedRectQD.left = obj->window.clipRect.left + npport->portx;
        revealedRectQD.bottom = obj->window.clipRect.bottom + npport->porty;
        revealedRectQD.right = obj->window.clipRect.right + npport->portx;
        ::SetOrigin(npport->portx, npport->porty);
        ::ClipRect(&revealedRectQD);

        return TRUE;
    }

    return FALSE;
}

static void endDrawQD(PluginObject *obj)
{
    if (!obj)
        return;
    ::SetOrigin(obj->savedPortLeftQD, obj->savedPortTopQD);
    ::SetClip(obj->savedClipQD);
    ::SetPort(obj->savedPortQD);
}

static void drawStringQD(char* text,
                         Rect drawRect)
{
    if (!text)
        return;

    char *buffer = (char *) malloc(strlen(text) + 1);
    if (!buffer)
        return;
    strcpy(buffer, text);

    char *token = strtok(buffer, "\n");
    if (!token) {
        free(buffer);
        return;
    }

    ::TextFont(1);
    ::TextFace(bold);
    ::TextMode(srcCopy);
    ::TextSize(12);

    FontInfo fontInfo;
    ::GetFontInfo(&fontInfo);

    short textHeight = fontInfo.ascent + fontInfo.descent + fontInfo.leading;

    short drawPosX = drawRect.left + 5;
    short drawPosY = drawRect.top + textHeight;

    while (token) {
        short length = strlen(token);
        ::MoveTo(drawPosX, drawPosY);
        ::DrawText(token, 0, length);
        drawPosY += textHeight;
        token = strtok(NULL, "\n");
    }

    free(buffer);
}

static void doDrawQD(PluginObject *obj)
{
    Rect drawRect;
    RGBColor black = { 0x0000, 0x0000, 0x0000 };
    RGBColor white = { 0xFFFF, 0xFFFF, 0xFFFF };
    RGBColor lightGray = { 0xA000, 0xA000, 0xA000 };

    char *pTheText = (char *) [obj->debugInfo UTF8String];

    drawRect.top = 0;
    drawRect.left = 0;
    drawRect.bottom = obj->window.height;
    drawRect.right = obj->window.width;

    ::PenNormal();
    ::RGBForeColor(&black);
    if (obj->hasFocus) {
        ::RGBBackColor(&white);
    } else {
        ::RGBBackColor(&lightGray);
    }

    Pattern qdPattern;
    ::GetQDGlobalsWhite(&qdPattern);
    ::FillRect(&drawRect, &qdPattern);

    if (obj->hasFocus)
        ::FrameRect(&drawRect);
    drawStringQD(pTheText, drawRect);
}

static void handleDraw(PluginObject *obj)
{
    if (obj->supportsCoreGraphics) {
        NSGraphicsContext *oldContext = [[NSGraphicsContext currentContext] retain];

        NSGraphicsContext *context =
            [NSGraphicsContext graphicsContextWithGraphicsPort:((NP_CGContext *)obj->window.window)->context
                                                       flipped:YES];

        [NSGraphicsContext setCurrentContext:context];

        NSRect rect = NSMakeRect(0, 0, obj->window.width, obj->window.height);

        if (obj->hasFocus) {
            [[NSColor whiteColor] set];
        } else {
            [[NSColor lightGrayColor] set];
        }
        [NSBezierPath fillRect:rect];

        // If the plugin has focus, draw a focus indicator
        if (obj->hasFocus) {
            [[NSColor blackColor] set];
            NSBezierPath *path = [NSBezierPath bezierPathWithRect:rect];
            [path setLineWidth:5];
            [path stroke];
        }

        [obj->debugInfo drawAtPoint:NSMakePoint(10, 10) withAttributes:nil];

        [NSGraphicsContext setCurrentContext:oldContext];
    } else {
        if (startDrawQD(obj)) {
            doDrawQD(obj);
            endDrawQD(obj);
        }
    }
}

static NSString *eventTypeCocoa(NPCocoaEventType type)
{
    switch (type) {
        case NPCocoaEventScrollWheel:
            return @"NPCocoaEventScrollWheel";
        case NPCocoaEventMouseDown:
            return @"NPCocoaEventMouseDown";
        case NPCocoaEventMouseUp:
            return @"NPCocoaEventMouseUp";            
        case NPCocoaEventMouseMoved:
            return @"NPCocoaEventMouseMoved";            
        case NPCocoaEventMouseDragged:
            return @"NPCocoaEventMouseDragged";            
        case NPCocoaEventMouseEntered:
            return @"NPCocoaEventMouseEntered";            
        case NPCocoaEventMouseExited:
            return @"NPCocoaEventMouseExited";
        case NPCocoaEventKeyDown:
            return @"NPCocoaEventKeyDown";
        case NPCocoaEventKeyUp:
            return @"NPCocoaEventKeyUp";
        case NPCocoaEventFlagsChanged:
            return @"NPCocoaEventFlagsChanged";
        default:
            return @"Unknown";
    }    
}

static NSString *eventTypeQD(EventKind type)
{
    switch (type) {
        case mouseDown:
            return @"mouseDown";
        case mouseUp:
            return @"mouseUp";
        case keyDown:
            return @"keyDown";
        case keyUp:
            return @"keyUp";
        case getFocusEvent:
            return @"getFocusEvent";
        case loseFocusEvent:
            return @"loseFocusEvent";
        default:
            return @"Unknown";
    }    
}

static void invalidatePlugin(PluginObject *obj)
{
    NPRect rect;
    rect.left = 0;
    rect.top = 0;
    rect.right = obj->window.width;
    rect.bottom = obj->window.height;
    
    browser.invalidaterect(obj->npp, &rect);    
}

static void handleMouseEvent(PluginObject *obj, void *event)
{
    NSString *debugInfo = nil;
    if (obj->supportsCocoa) {
        NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8x\n"
                                               "Coordinates: (%g, %g)\n"
                                               "Button number: %d\n"
                                               "Click count: %d\n"
                                               "Delta: (%g, %g, %g)\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                               eventTypeCocoa(cocoaEvent->type), 
                                               cocoaEvent->event.mouse.modifierFlags,
                                               cocoaEvent->event.mouse.pluginX,
                                                   cocoaEvent->event.mouse.pluginY,
                                               cocoaEvent->event.mouse.buttonNumber,
                                               cocoaEvent->event.mouse.clickCount,
                                               cocoaEvent->event.mouse.deltaX,
                                                   cocoaEvent->event.mouse.deltaY,
                                                   cocoaEvent->event.mouse.deltaZ,
                                               (cocoaEvent->event.mouse.modifierFlags &
                                                   NSAlternateKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.mouse.modifierFlags &
                                                   NSControlKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.mouse.modifierFlags &
                                                   NSCommandKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.mouse.modifierFlags &
                                                   NSShiftKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.mouse.modifierFlags &
                                                   NSAlphaShiftKeyMask) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type: %@"
                                ", Modifiers: 0x%.8x"
                                ", Coordinates: (%g, %g)"
                                ", Button number: %d"
                                ", Click count: %d"
                                ", Delta: (%g, %g, %g)",
                                (unsigned) obj->npp,
                                eventTypeCocoa(cocoaEvent->type), 
                                cocoaEvent->event.mouse.modifierFlags,
                                cocoaEvent->event.mouse.pluginX,
                                    cocoaEvent->event.mouse.pluginY,
                                cocoaEvent->event.mouse.buttonNumber,
                                cocoaEvent->event.mouse.clickCount,
                                cocoaEvent->event.mouse.deltaX,
                                    cocoaEvent->event.mouse.deltaY,
                                    cocoaEvent->event.mouse.deltaZ);
    } else {
        EventRecord *qdEvent = (EventRecord *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8X\n"
                                               "Coordinates: (%d, %d)\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                               eventTypeQD(qdEvent->what),
                                               qdEvent->modifiers,
                                               qdEvent->where.h, qdEvent->where.v,
                                               ((qdEvent->modifiers & optionKey) ||
                                                   (qdEvent->modifiers & rightOptionKey))
                                                   ? "Down" : "Up",
                                               ((qdEvent->modifiers & controlKey) ||
                                                   (qdEvent->modifiers & rightControlKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & cmdKey) ? "Down" : "Up",
                                               ((qdEvent->modifiers & shiftKey) ||
                                                   (qdEvent->modifiers & rightShiftKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & alphaLock) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type: %@"
                                ", Modifiers: 0x%.8X"
                                ", Coordinates: (%d, %d)",
                                (unsigned) obj->npp,
                                eventTypeQD(qdEvent->what),
                                qdEvent->modifiers,
                                qdEvent->where.h, qdEvent->where.v);
    }
    [obj->debugInfo release];
    obj->debugInfo = [debugInfo retain];
 
    invalidatePlugin(obj);
}

static void handleKeyboardEvent(PluginObject *obj, void *event)
{
    NSString *debugInfo = nil;
    if (obj->supportsCocoa) {
        NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8x\n"
                                               "Characters: %@\n"
                                               "Character code: %d\n"
                                               "Characters ignoring modifiers: %@\n"
                                               "Is a repeat: %@\n"
                                               "Key code: %d\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                                eventTypeCocoa(cocoaEvent->type), 
                                                cocoaEvent->event.key.modifierFlags,
                                                cocoaEvent->event.key.characters,
                                                [(NSString *)cocoaEvent->event.key.characters characterAtIndex:0],
                                                cocoaEvent->event.key.charactersIgnoringModifiers,
                                                cocoaEvent->event.key.isARepeat ? @"YES" : @"NO",
                                                cocoaEvent->event.key.keyCode,
                                               (cocoaEvent->event.key.modifierFlags &
                                                   NSAlternateKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.key.modifierFlags &
                                                   NSControlKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.key.modifierFlags &
                                                   NSCommandKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.key.modifierFlags &
                                                   NSShiftKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->event.key.modifierFlags &
                                                   NSAlphaShiftKeyMask) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type %@"
                                ", Modifiers: 0x%.8x"
                                ", Characters: %@"
                                ", Character code: %d"
                                ", Characters ignoring modifiers: %@"
                                ", Is a repeat: %@"
                                ", Key code: %d",
                                (unsigned) obj->npp,
                                eventTypeCocoa(cocoaEvent->type), 
                                cocoaEvent->event.key.modifierFlags,
                                cocoaEvent->event.key.characters,
                                [(NSString *)cocoaEvent->event.key.characters characterAtIndex:0],
                                cocoaEvent->event.key.charactersIgnoringModifiers,
                                cocoaEvent->event.key.isARepeat ? @"YES" : @"NO",
                                cocoaEvent->event.key.keyCode);
    } else {
        EventRecord *qdEvent = (EventRecord *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8X\n"
                                               "Character code: %d\n"
                                               "Key code: %d\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                               eventTypeQD(qdEvent->what),
                                               qdEvent->modifiers,
                                               qdEvent->message & charCodeMask,
                                               (qdEvent->message & keyCodeMask) >> 8,
                                               ((qdEvent->modifiers & optionKey) ||
                                                   (qdEvent->modifiers & rightOptionKey))
                                                   ? "Down" : "Up",
                                               ((qdEvent->modifiers & controlKey) ||
                                                   (qdEvent->modifiers & rightControlKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & cmdKey) ? "Down" : "Up",
                                               ((qdEvent->modifiers & shiftKey) ||
                                                   (qdEvent->modifiers & rightShiftKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & alphaLock) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type %@"
                                ", Modifiers: 0x%.8X"
                                ", Character code: %d"
                                ", Key code: %d",
                                (unsigned) obj->npp,
                                eventTypeQD(qdEvent->what),
                                qdEvent->modifiers,
                                qdEvent->message & charCodeMask,
                                (qdEvent->message & keyCodeMask) >> 8);
    }

    [obj->debugInfo release];
    obj->debugInfo = [debugInfo retain];

    invalidatePlugin(obj);
}

static void handleGetFocusEvent(PluginObject *obj)
{
    if (!obj)
        return;

    if (obj->hasFocus) {
        NSLog(@"DebugEventsPlugin (0x%.8X): Redundant getFocus event", (unsigned) obj->npp);
    } else {
        NSLog(@"DebugEventsPlugin (0x%.8X): getFocus", (unsigned) obj->npp);
        obj->hasFocus = TRUE;
        invalidatePlugin(obj);
    }
}

static void handleLoseFocusEvent(PluginObject *obj)
{
    if (!obj)
        return;

    if (!obj->hasFocus) {
        NSLog(@"DebugEventsPlugin (0x%.8X): Redundant loseFocus event", (unsigned) obj->npp);
    } else {
        NSLog(@"DebugEventsPlugin (0x%.8X): loseFocus", (unsigned) obj->npp);
        obj->hasFocus = FALSE;
        invalidatePlugin(obj);
    }
}

int16 NPP_HandleEvent(NPP instance, void* event)
{
    if (!checkInstance(instance)) {
        NSLog(@"DebugEventsPlugin: NPP_HandleEvent() error return");
        return 0;
    }
    if (!event) {
        NSLog(@"DebugEventsPlugin: NPP_HandleEvent(): NULL 'event'!");
        return 0;
    }

    PluginObject *obj = (PluginObject *) instance->pdata;

    if (obj->supportsCocoa) {
        NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;
        switch(cocoaEvent->type) {
            case NPCocoaEventFocusChanged:
                if (cocoaEvent->event.focus.hasFocus) {
                    handleGetFocusEvent(obj);
                } else {
                    handleLoseFocusEvent(obj);
                }
                return 1;

            case NPCocoaEventDrawRect:
                handleDraw(obj);
                return 1;

            case NPCocoaEventKeyDown:
            case NPCocoaEventKeyUp:
            case NPCocoaEventFlagsChanged:
                handleKeyboardEvent(obj, event);
                return 1;

            case NPCocoaEventMouseDown:
            case NPCocoaEventMouseUp:

            // FIXME: NPCocoaEventMouseMoved is currently disabled in order to see other events more clearly
            // without "drowning" in mouse moved events.
            //case NPCocoaEventMouseMoved:
            case NPCocoaEventMouseEntered:
            case NPCocoaEventMouseExited:
            case NPCocoaEventMouseDragged:
            case NPCocoaEventScrollWheel:
                handleMouseEvent(obj, event);
                return 1;
            default:
                break;
        }
    } else {
        EventRecord *qdEvent = (EventRecord *) event;
        switch (qdEvent->what) {
            case mouseDown:
            case mouseUp:
                handleMouseEvent(obj, event);
                return 1;
            case keyDown:
            case keyUp:
                handleKeyboardEvent(obj, event);
                return 1;
            case updateEvt:
                handleDraw(obj);
                return 1;
            case getFocusEvent:
                handleGetFocusEvent(obj);
                return 1;
            case loseFocusEvent:
                handleLoseFocusEvent(obj);
                return 1;
            default:
                break;
        }
    }
    
    return 0;
}

void NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
    NSLog(@"DebugEventsPlugin: NPP_URLNotify(): instance 0x%.8X, url %s",
           (unsigned) instance, url);
    checkInstance(instance);
}

NPError NPP_GetValue(NPP instance, NPPVariable variable, void *value)
{
    NSLog(@"DebugEventsPlugin: NPP_GetValue(): instance 0x%.8X, variable %i",
           (unsigned) instance, variable);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;
    return NPERR_GENERIC_ERROR;
}

NPError NPP_SetValue(NPP instance, NPNVariable variable, void *value)
{
    NSLog(@"DebugEventsPlugin: NPP_SetValue(): instance 0x%.8X, variable %i",
           (unsigned) instance, variable);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;
    return NPERR_GENERIC_ERROR;
}
